/*
    CFS0.11 filesystem implementation for FUSE
    Copyright (C) 2005-2020 Kajtar Zsolt (Soci/Singular) <soci@c64.rulez.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/
#include "block.h"

#define _LARGEFILE64_SOURCE

#ifdef linux
/* For pread()/pwrite() */
#define _XOPEN_SOURCE 500
#endif

#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>

#ifndef O_LARGEFILE
#define O_LARGEFILE 0
#endif

static int f1 = -1;
static unsigned int geometry_cylinder; //constant
static unsigned int geometry_head; //constant
static unsigned int geometry_sector; //constant

const char *block_init(const char *name) {
    f1 = open(name, O_RDWR | O_LARGEFILE);
    return (f1 >= 0) ? NULL : strerror(errno);
}

void block_destroy(void) {
    if (f1 < 0) return;
    fsync(f1);
    close(f1);
    f1 = -1;
}

int block_write(const Block *ezt, Offset ide) {
    if (ide == null_sector) return -EIO;
//    fprintf(stderr, "WRITE (%u)\n", ide);
#ifdef __CYGWIN__
    if (lseek(f1, (off_t)ide << 9, SEEK_SET) < 0 || write(f1, ezt->data, 512) < 0) return errno;
#else
    if (pwrite(f1, ezt->data, 512, (off_t)ide << 9) < 0) return errno;
#endif
    return 0;
}

int block_read(Block *ezt, Offset ide) {
    ssize_t ret;
    if (ide == null_sector) {
        memset(ezt->data, 0, 512);
        return -EIO;
    }
//    fprintf(stderr, "READ (%u)\n", ide);
#ifdef __CYGWIN__
    if (lseek(f1, (off_t)ide << 9, SEEK_SET) == (off_t)-1) {
        memset(ezt->data, 0, 512);
        ret = -1;
    } else ret = read(f1, ezt->data, 512);
#else
    ret = pread(f1, ezt->data, 512, (off_t)ide << 9);
#endif
    if (ret < 0) return errno;
    if (ret < 512) memset(ezt->data + ret, 0, 512 - ret);
    return 0;
}

int offset_geometry(const unsigned char ez[]) {
    geometry_head = (ez[0] & 0x4f) + 1;
    geometry_cylinder = (ez[1] << 8) + ez[2];
    geometry_sector = ez[3];
    return !(geometry_head & 0x40) && (geometry_head == 0 || geometry_sector == 0);
}

Offset offset_from(const unsigned char ez[]) {
    Offset cylinder, head, sector;
    head = ez[0] & 0x0f;
    cylinder = (ez[1] << 8) + ez[2];
    sector = ez[3];
    if (ez[0] & 0x40) {
        return sector + (cylinder << 8) + (head << 24);
    }
    if (sector == 0 || sector > geometry_sector || head >= geometry_head || cylinder > geometry_cylinder) return null_sector;
    return (cylinder * geometry_head + head) * geometry_sector + sector - 1;
}

void offset_to(unsigned char ide[], Offset ezt) {
    if (geometry_head & 0x40) {
        ide[0] = ((ezt >> 24) & 0x0f) | 0x40;
        ide[1] = ezt >> 16;
        ide[2] = ezt >> 8;
        ide[3] = ezt;
        return;
    }
    ide[3] = (ezt % geometry_sector) + 1;
    ezt /= geometry_sector;
    ide[0] = ezt % geometry_head;
    ezt /= geometry_head;
    ide[1] = ezt >> 8;
    ide[2] = ezt;
}

